import os
import warnings
from collections import defaultdict
from typing import Type, Union, NoReturn, Any, TypeVar

# noinspection PyProtectedMember
from aflowey.single_executor import _exec
from fhir.resources.identifier import Identifier
from fhir.resources.reference import Reference
from fhir.resources.resource import Resource
from fhirpy.base import AsyncResource
from fhirpy.base.utils import AttrDict
from fhirpy.lib import AsyncFHIRResource, AsyncFHIRReference
from seito.monad.try_ import try_

from smart_on_fhir_client.client import (
    SmartOnFhirClientBuilder,
    SmartOnFhirClient,
    CustomFHIRSearchSet,
)
from smart_on_fhir_client.partner import Partner, TargetUrlStrategy, Organization
from smart_on_fhir_client.requester.fhir_resource import CustomFHIRResource


class SearchSet:
    """
    Wrapper around fhir client class to perform auto conversion
    to pydantic model if needed
    """

    def __init__(
        self,
        search: CustomFHIRSearchSet,
        fhir_manager: lambda: FhirContextManager,
        client: SmartOnFhirClient,
    ):
        self._search = search
        self._fhir_manager = fhir_manager
        self._client = client

    @staticmethod
    def _process_result(result, return_as: Type):
        if result is None or not result:
            return result
        result_is_list = isinstance(result, list)
        if return_as is not None:
            if result_is_list:
                return [return_as(**res) for res in result]
            return return_as(**result)
        return result

    def limit(self, value):
        return SearchSet(self._search.limit(value), self._fhir_manager, self._client)

    def sort(self, value):
        return SearchSet(self._search.sort(value), self._fhir_manager, self._client)

    def revinclude(self, *args, **kwargs):
        return SearchSet(
            self._search.revinclude(*args, **kwargs), self._fhir_manager, self._client
        )

    def include(self, *args, **kwargs):
        return SearchSet(
            self._search.include(*args, **kwargs), self._fhir_manager, self._client
        )

    async def fetch_raw(self, return_as=None):
        result = await self._search.fetch_raw()
        return self._process_result(result, return_as=return_as)

    async def fetch(self, all_resources: bool = False, return_as=None):
        # maybe wrap in attempt
        result = await self._search.fetch(all_resources=all_resources)
        return self._process_result(result, return_as=return_as)

    async def post_fetch(self, return_as=None, enable_modifier: bool = False):
        result = await self._search.post_fetch(enable_modifier=enable_modifier)
        return self._process_result(result, return_as=return_as)

    async def first(self, return_as=None):
        """return first instance converted to the target class"""
        result = await self._search.first()
        return self._process_result(result, return_as=return_as)

    async def post_first(self, return_as=None, enable_modifier: bool = False):
        result = await self._search.post_first(enable_modifier=enable_modifier)
        return self._process_result(result, return_as=return_as)


class ClientProxy:
    def __init__(
        self, _id: str, client: SmartOnFhirClient, fhir_manager: "FhirContextManager"
    ) -> None:
        # id of the resource (Patient, Organisation, Practitioner...)
        self._id = _id
        # let access to the raw fhir api client
        self.client = client
        # add a manager
        self._fhir_manager = fhir_manager
        # allow research stuff
        self._target = self.client.resources(_id)

    def create_resource(
        self, data: Resource | AsyncResource | AttrDict
    ) -> CustomFHIRResource:
        wanted_cls = self._fhir_manager.cls_by_partner_id[self.client.client_name].get(
            self._id
        )
        cls = wanted_cls or CustomFHIRResource
        data = data.dict() if isinstance(data, Resource) else data
        return cls(
            self.client,
            self._id,
            **data,
        )

    def search(self, **kwargs) -> SearchSet:
        return SearchSet(self._target.search(**kwargs), self._fhir_manager, self.client)

    def reference(self, the_id: str) -> AsyncFHIRReference:
        return self.client.reference(self._id, the_id)

    async def by_id(self, the_id: str) -> CustomFHIRResource:
        return await self.reference(the_id).to_resource()

    async def save(
        self,
        resource: Resource | AsyncFHIRResource | CustomFHIRResource,
        **kwargs,
    ):

        resource_to_save = (
            self._fhir_manager.create_async_fhir_resource(
                self.client, resource, **kwargs
            )
            if isinstance(resource, Resource)
            else resource
        )
        # update in place
        await try_(resource_to_save.save)().or_raise(ValueError("Error"))

        # useful to get the hook cls by partner
        return resource_to_save

    async def update(
        self,
        resource: Resource | AsyncResource | CustomFHIRResource,
        **kwargs: Any,
    ):
        by_alias = kwargs.pop("by_alias")
        resource_to_save = self._fhir_manager.create_async_fhir_resource(
            self.client, resource, **kwargs
        )

        return await try_(resource_to_save.update)(
            **kwargs, by_alias=by_alias
        ).or_raise()

    def upsert(self, resource):
        ...

    async def delete(
        self, resource: Resource | AsyncResource | CustomFHIRResource, **kwargs
    ):
        resource_to_save = self._fhir_manager.create_async_fhir_resource(
            self.client, resource, **kwargs
        )
        return await try_(resource_to_save.delete)().or_raise()


T = TypeVar("T", bound=Resource)


class FhirContextRequester:
    """
    Fhir requester attached to one tenant
    """

    # to be extended
    RESOURCES = frozenset(
        {
            "Patient",
            "Organization",
            "Practitioner",
            "Condition",
            "ResearchStudy",
            "ResearchSubject",
            "Medication",
            "MedicationAdministration",
            "MedicationStatement",
            "MedicationRequest",
            "Encounter",  # 🔔 lifen !
            "CareTeam",
            "PractitionerRole",
            "List",
            "QuestionnaireResponse",
            "Communication",
            "CommunicationRequest",
            "Provenance",
        }
    )

    def __init__(self, client):
        self._id = client.url.split("/")[-1]
        self._client = client
        self._fhir_manager = client.fhir_manager

        for resource_name in FhirContextRequester.RESOURCES:
            self.__setattr__(
                resource_name, ClientProxy(resource_name, client, self._fhir_manager)
            )

    @staticmethod
    def _get_result_as_or_raw(
        resource: AsyncResource, *, return_as: Type[T] = None
    ) -> AsyncResource | T:
        if return_as:
            return return_as(**resource)
        return resource

    async def resolve_ref(
        self,
        reference: Union[Reference, str],
        *,
        return_as: Type[T] = None,
        raise_if_none: bool = False,
    ) -> CustomFHIRResource | T | None | NoReturn:
        """resolve a fhir reference"""
        if not reference:
            if raise_if_none:
                raise ValueError("Reference is None")
            return None

        is_fhir_reference = isinstance(reference, Reference)
        if (
            is_fhir_reference
            and reference.reference is None
            and reference.identifier is not None
            and reference.type is not None
        ):
            # handle identifier
            # noinspection PyTypeChecker
            identifier_as_fhir: Identifier = reference.identifier
            result = (
                await self._client.resources(reference.type)
                .search(identifier=identifier_as_fhir.value)
                .first()
            )
            return self._get_result_as_or_raw(result, return_as=return_as)

        if is_fhir_reference and reference.reference is not None:
            reference = reference.reference
        # used for conversion
        fhirpy_resource_dict = await self._client.reference(
            reference=reference
        ).to_resource()
        return self._get_result_as_or_raw(fhirpy_resource_dict, return_as=return_as)

    @property
    def patient(self) -> ClientProxy:
        return getattr(self, "Patient")

    @property
    def organization(self) -> ClientProxy:
        return getattr(self, "Organization")

    @property
    def practitioner(self) -> ClientProxy:
        return getattr(self, "Practitioner")

    @property
    def condition(self) -> ClientProxy:
        return getattr(self, "Condition")

    @property
    def research_study(self) -> ClientProxy:
        return getattr(self, "ResearchStudy")

    @property
    def research_subject(self) -> ClientProxy:
        return getattr(self, "ResearchSubject")

    @property
    def medication(self) -> ClientProxy:
        return getattr(self, "Medication")

    @property
    def medication_administration(self) -> ClientProxy:
        return getattr(self, "MedicationAdministration")

    @property
    def medication_statement(self) -> ClientProxy:
        return getattr(self, "MedicationStatement")

    @property
    def medication_request(self) -> ClientProxy:
        return getattr(self, "MedicationRequest")

    @property
    def encounter(self) -> ClientProxy:
        # 🔔 lifen !
        return getattr(self, "Encounter")

    @property
    def care_team(self) -> ClientProxy:
        return getattr(self, "CareTeam")

    @property
    def practitioner_role(self) -> ClientProxy:
        return getattr(self, "PractitionerRole")

    @property
    def list(self) -> ClientProxy:
        return getattr(self, "List")

    @property
    def questionnaire_response(self) -> ClientProxy:
        return getattr(self, "QuestionnaireResponse")

    @property
    def communication(self) -> ClientProxy:
        return getattr(self, "Communication")

    @property
    def communication_request(self) -> ClientProxy:
        return getattr(self, "CommunicationRequest")

    @property
    def provenance(self) -> ClientProxy:
        return getattr(self, "Provenance")


class FhirContextManager:
    """

    main context manager to perform request to the fhir server
    allow handling CRUD operation. This is finally a small wrapper around
    the client API.

    """

    OWN_FHIR_URL = os.getenv("OWN_FHIR_URL", "http://localhost:8080/fhir")

    def __init__(self, own_fhir_url: str | None = None):
        self.OWN_FHIR_URL = own_fhir_url or self.OWN_FHIR_URL
        self.cls_by_partner_id = defaultdict(dict)

    def set_own_fhir_url(self, url: str):
        self.OWN_FHIR_URL = url
        return self

    def create_async_fhir_resource(
        self,
        client: SmartOnFhirClient,
        resource: CustomFHIRResource | AsyncResource | Resource,
        **kwargs: Any,
    ) -> CustomFHIRResource | NoReturn:

        client_name = client.client_name
        resource_type = resource.get("resource_type") or resource.get("resourceType")
        wanted_cls = self.cls_by_partner_id[client_name].get(resource_type)
        cls = wanted_cls or CustomFHIRResource
        match resource:
            case Resource():
                return cls(
                    client,
                    resource.resource_type,
                    **resource.dict(by_alias=kwargs.get("by_alias", False)),
                )
            case AsyncResource():
                return cls(client, resource.resourceType, **resource)
            case CustomFHIRResource():
                return resource
            case _:
                raise ValueError("Could not create async fhir resource")

    @staticmethod
    def _get_tenant_id(
        target_url_strategy: TargetUrlStrategy,
        partner_name: str,
        client_name: str,
        organization: Organization,
    ) -> str:
        match target_url_strategy:
            case TargetUrlStrategy.NONE:
                return ""
            case TargetUrlStrategy.PARTNER:
                return partner_name
            case TargetUrlStrategy.ORGANIZATION_NAME:
                return client_name
            case TargetUrlStrategy.CUSTOM:
                return organization.parameters["target_tenant"]
        raise ValueError("Invalid target url strategy")

    def register_partner(
        self,
        client_name: str,
        partner: Partner,
        client: SmartOnFhirClient,
        organization: Organization,
        target_server_authorization: str = None,
    ) -> None:
        """Add a partner requester with the partition of the partner"""
        partner_name = partner.name
        self.__setattr__(client_name, FhirContextRequester(client))

        target_url_strategy = (
            organization.target_url_strategy
            if organization
            else TargetUrlStrategy.PARTNER
        )

        tenant_id = self._get_tenant_id(
            target_url_strategy, partner_name, client_name, organization
        )
        target_url = (
            self.OWN_FHIR_URL if not tenant_id else f"{self.OWN_FHIR_URL}/{tenant_id}"
        )
        self.__setattr__(
            f"TARGET_{client_name}",
            FhirContextRequester(
                SmartOnFhirClient(
                    url=target_url,
                    authorization=f"Bearer {target_server_authorization}",
                    partner=partner,
                    fhir_manager=self,
                )
            ),
        )

    async def register_fhir_client_async(
        self,
        builder: SmartOnFhirClientBuilder,
    ):
        fhir_client = await builder.build(self)

        # unpacking partner information
        partner = builder.partner
        partner_name = partner.name

        organization = builder.organization
        # unpacking organization information
        organization_name = organization.slug if organization else ""

        # the final client name is the organization if it exists else the partner name
        client_name = organization_name or partner_name

        # register for each resource, its own callback
        for resource_type, cb in builder.cls_by_resource.items():
            self.cls_by_partner_id[client_name][resource_type] = cb

        target_server_authorization = builder.target_fhir_server_authorization or ""

        if callable(target_server_authorization):
            target_server_authorization = await _exec(target_server_authorization) or ""

        self.register_partner(
            client_name,
            partner,
            fhir_client,
            organization,
            target_server_authorization,
        )

    def req(self, client_name) -> FhirContextRequester | None:
        partner_requester = getattr(self, client_name)
        if partner_requester is None:
            warnings.warn(
                f"fhir manager does not have a '{client_name}' partner registered. Did you registered it ?"
            )
        return partner_requester


# create a singleton of the fhir manager
default_fhir_client_manager = FhirContextManager()
fhir_client_manager = default_fhir_client_manager
